from uuid import UUID
from app.core.model import Schema


class User(Schema):
    uid: UUID
    email: str


class ParsedToken(Schema):
    user: User
    token: str  # token string for downstream use
